/****************************************************************************/
/* general.h - zakladni spolecne definice pro praci s WinBase602            */
/* (C) Janus Drozd, 1992, 93, 94, 95, 96, 97, 98, 99, 2000                  */
/* verze: 6.0, 6.1 (32-bit)                                                 */
/****************************************************************************/
#ifndef __GENERAL_H__
#define __GENERAL_H__

#ifndef GENERAL_DEF // neni casti obecnejsich deklaraci //////////////////////
#ifndef __WIN32__
#define __WIN32__
#endif

#ifndef WIN32
#define WIN32
#endif

#define DllImport   __declspec(dllimport)
#define DllPrezen   __declspec(dllimport)
#define DllKernel   __declspec(dllimport)
#define DllViewed   __declspec(dllimport)
#define DllExport   __declspec(dllexport)
#define DllInstserv __declspec(dllimport)
#define DllWbed     __declspec(dllimport)
#define oexport
#endif // GENERAL_DEF neni definovano ////////////////////////////////////////

#define _near
#define __near
#define huge
#define _huge
#define __huge

#ifdef __WATCOMC__
#pragma off (unreferenced)
#define EMPTY_ARRAY_INDEX
#else
#define EMPTY_ARRAY_INDEX   0
#endif /* __WATCOMC__ */

#ifndef NULL
#define NULL  0
#endif

#define CFNC extern "C"

typedef unsigned long  uns32;
typedef signed   long  sig32;
typedef unsigned short uns16;
typedef signed   short sig16;
typedef unsigned char  uns8;
typedef signed   char  sig8;
typedef unsigned char  byte;

typedef       char *  tptr;
typedef const char * ctptr;

typedef uns8     wbbool;
#define wbfalse  (wbbool)0
#define wbtrue   (wbbool)1

/* Zakladni typy databazoveho jadra */
typedef sig16  ttablenum;
typedef sig16  tobjnum;
typedef sig16  tcursnum;
typedef sig16  tcurstab;
typedef uns32  trecnum;
typedef uns8   tattrib;
typedef uns8   tcateg;
typedef uns8   tright;     /* zastaraly typ */
typedef uns16  tdright;    /* zastaraly typ */

#define OBJ_NAME_LEN  31   /* max. delka jmena databazovych objektu */
typedef char tobjname[OBJ_NAME_LEN+1];
#define NAMELEN       18   /* delka jmena identifikatoru */
typedef char tname[NAMELEN+1];
#define UUID_SIZE     12   /* delka identifikace objektu */
typedef uns8 WBUUID[UUID_SIZE];

typedef HWND window_id;
typedef enum { OPER_SET=0, OPER_GET=1, OPER_GETEFF=2, OPER_SETREPL=3 } t_oper;
typedef enum { VT_OBJNUM=0,  VT_NAME=1, VT_UUID=2, VT_NAME3=3 } t_valtype;
typedef enum { READ_UNCOMMITTED, READ_COMMITTED, REPEATABLE_READ, SERIALIZABLE } t_isolation;

#define NOOBJECT  (tobjnum)-1   // cislo neexistujiciho objektu
#define NOCURSOR  (tcursnum)-1  // cislo neexistujiciho kurzoru
#define NOINDEX   0xffff        // hodnota parametru index, nejde-li o multiatribut 
#define NORECNUM  (trecnum)-1   // cislo neexistujiciho zaznamu
#define NOATTRIB  (tattrib)-1   // cislo neexistujiciho sloupce

#define MAX_PACKAGED_REQS    25  /* max. pocet pozadavku v baliku */

/************************* Cisla systemovych tabulek ************************/
#define TAB_TABLENUM   0   /* cislo tabulky tabulek */
#define OBJ_TABLENUM   1   /* cislo tabulky objektu */
#define USER_TABLENUM  2   /* cislo tabulky uzivatelu */
#define SRV_TABLENUM   3   /* cislo tabulky serveru */
#define REPL_TABLENUM  4   /* cislo tabulky replikacnich pravidel */
#define KEY_TABLENUM   5   /* cislo tabulky verejnych klicu */

/**************************** Atribut "DELETED": ****************************/
#define DEL_ATTR_NUM   0   /* cislo atributu "DELETED" */
#define NOT_DELETED    0   /* hodnoty atributu: zaznam neni zrusen */
#define DELETED        1   /*                   zaznam je zrusen   */
#define RECORD_EMPTY   2   /*                   zaznam je uvolnen  */

/*************** Atributy systemovych tabulek TABTAB A OBJTAB ***************/
#define OBJ_NAME_ATR   3 /* Jmeno objektu, STRING delky OBJ_NAME_LEN znaku */
#define OBJ_CATEG_ATR  4 /* Kategorie objektu, atribut typu CHAR */
#define APPL_ID_ATR    5 /* Id aplikace, k niz objekt patri, BINARY delky 12 bajtu */
#define OBJ_DEF_ATR    6 /* Definice objektu, atribut typu Text */
#define OBJ_FLAGS_ATR  7 /* Priznaky objektu, atribut typu SHORT */

/******************************* Kategorie: *********************************/
#define CATEG_TABLE        0  // tabulka
#define CATEG_USER         1  // uzivatel
#define CATEG_VIEW         2  // pohled
#define CATEG_CURSOR       3  // dotaz
#define CATEG_PGMSRC       4  // text programu
#define CATEG_PGMEXE       5  // prelozeny program
#define CATEG_MENU         6  // menu
#define CATEG_APPL         7  // aplikace
#define CATEG_PICT         8  // obrazek
#define CATEG_GROUP        9  // skupina
#define CATEG_ROLE        10  // role
#define CATEG_CONNECTION  11  // ODBC spojeni
#define CATEG_RELATION    12  // relace
#define CATEG_DRAWING     13  // nakres
#define CATEG_GRAPH       14  // graf
#define CATEG_REPLREL     15  // replikacni vztah
#define CATEG_PROC        16  // rutina
#define CATEG_TRIGGER     17  // trigger
#define CATEG_WWW         18  // WWW objekt
#define CATEG_FOLDER      19  // slozka
#define CATEG_SEQ         20  // sekvence
#define CATEG_KEY         25  // verejny klic uzivatele
#define CATEG_SERVER      26  // replikacni server
#define CATEG_COUNT       27
#define CATEG_MASK    0x7f
#define IS_LINK       0x80    // priznak spojovaciho objektu
#define CATEG_DIRCUR      27  // otevreny kurzor

/*************************** struktura kernel_info **************************/
typedef struct {
  char  version[6];   /* ASCIIZ oznaceni verze serveru */
  uns16 logged;       /* pocet prihlasenych uzivatelu */
  uns16 blocksize;    /* velikost clusteru */
  uns32 freeblocks;   /* pocet volnych clusteru v alokacni tabulce */
  uns16 frames;       /* pocet pametovych ramu */
  uns8  fixed_pages;  /* pocet ramu s fixovanym obsahem */
  uns8  max_users;    /* maximalni pocet uzivatelu */
  uns8  rtp;          /* provozni parametry */
  uns32 diskspace;    /* mnozstvi volneho mista na disku */
  tobjname server_name;     /* jmeno serveru */
  uns32 local_free_memory;  /* volna pamet u klienta */
  uns32 remote_free_memory; /* volna pamet na serveru */
  BOOL  networking;         /* sitova prace klienta */
  uns16 owned_cursors;      /* pocet otevrenych kurzoru klientem */
               } kernel_info;

/************** Hodnoty slozky "modtype" v zaznamu "modifrec" ***************/
#define MODSTOP     0
#define MODLEN      2
#define MODIND      3
#define MODINT      4
#define MODPTR      5
#define MODINDPTR   6

typedef struct {
  uns8 modtype;
  union umoddef {
    struct smodstop{ uns16 dummy; }              modstop;
    struct smodlen { uns16 dummy; }              modlen;
    struct smodind { uns16 index; }              modind;
    struct smodint { uns32 start;  uns16 size; } modint;
    struct smodp   { uns8  attr;  }              modptr;
    struct smodindp{ uns16 index;  uns8 attr; }  modindptr;
                } moddef;
               } modifrec;

/************************** hodnota typu money ******************************/
typedef struct
{ uns16   money_lo2;
  sig32   money_hi4;
} monstr;
/************************* Pristupova prava: ********************************/
#define RIGHT_READ       0x01  // calculated priviledge, never granted explicitly
#define RIGHT_WRITE      0x02  // calculated priviledge, never granted explicitly
#define RIGHT_APPEND     0x04  /* pravo vkladat zaznamy                     */
#define RIGHT_INSERT     0x04  /* dtto                                      */
#define RIGHT_DEL        0x08  /* pravo rusit zaznamy                       */
#define RIGHT_NEW_READ   0x10  /* pridelovat k novym zaznamum pravo cist    */
#define RIGHT_NEW_WRITE  0x20  /* pridelovat k novym zaznamum pravo prepsat */
#define RIGHT_NEW_DEL    0x40  /* pridelovat k novym zaznamum pravo zrusit  */
#define RIGHT_GRANT      0x80  /* universalni pravo poskytovat sva prava    */

#define PRIVIL_DESCR_SIZE     (1+64) // velikost zaznamu o pravech 

#define HAS_READ_PRIVIL(priv_val, atr) (*priv_val & RIGHT_READ  ||\
  priv_val[1+(atr-1) / 4] &  (1 << (2*((atr-1)%4))  ))
#define HAS_WRITE_PRIVIL(priv_val,atr) (*priv_val & RIGHT_WRITE ||\
  priv_val[1+(atr-1) / 4] &  (1 << ((2*((atr-1)%4))+1)))
#define SET_READ_PRIVIL(priv_val, atr)\
  priv_val[1+(atr-1) / 4] |= (1 << (2*((atr-1)%4))  )
#define SET_WRITE_PRIVIL(priv_val,atr)\
  priv_val[1+(atr-1) / 4] |= (1 << ((2*((atr-1)%4))+1))

/********************* Hodnoty "NONE" ruznych typu: *************************/
#define NONEBOOLEAN   ((uns8)0x80)
#define NONECHAR      ((uns8)0)
#define NONEDATE      0x80000000L
#define NONETIME      0x80000000L
#define NONETIMESTAMP 0x80000000L
#define NONEMONEY     /* nelze takto definovat, ma hodnotu 0,0,0,0,0,0x80 */
#define NONEINTEGER   ((sig32)0x80000000L)
#define NONESHORT     ((sig16)0x8000)
#define NONEREAL      -1.7001e308
#define NONESTRING    ""
#define NONEPTR       ((uns32)-1)

/********************* Parametry funkce Set_sql_option **********************/
#define SQLOPT_NULLEQNULL        1 // hodnota NULL se rovna NULL
#define SQLOPT_NULLCOMP          2 // NULL se porovnavat s ne-NULL hodnotami
#define SQLOPT_RD_PRIVIL_VIOL    4 // hodnota bez prava cteni se jevi najo NULL
#define SQLOPT_MASK_NUM_RANGE    8 // preteceni ciselneho typu pri konverzi vrati NULL
#define SQLOPT_MASK_INV_CHAR    16 // nekonvertovatelny retezec znaku vrati NULL
#define SQLOPT_MASK_RIGHT_TRUNC 32 // zkraceni retezce pri konverzi se maskuje
#define SQLOPT_EXPLIC_FREE      64 // zrusene zaznamy se uvolni az pri explicitnim Free_deleted
#define SQLOPT_OLD_ALTER_TABLE 128 // specificke ALTER TABLE

/******************************* Cisla chyb: ********************************/
#ifndef GENERAL_DEF // neni casti obecnejsich deklaraci //////////////////////
#ifndef NO_ERROR    /* conflict with winerror.h */
#define NO_ERROR                 0
#endif
#undef  IS_ERROR
#define IS_ERROR                 128
#endif

#define ANS_OK                   0
#define NOT_ANSWERED             0xff
#define BAD_MODIF                0x80
#define NO_RIGHT                 0x81
#define BAD_ELEM_NUM             0x82
#define OUT_OF_TABLE             0x83 // 02000
#define TABLE_FULL               0x84
#define CURSOR_MISUSE            0x85
#define BAD_OPCODE               0x86
#define CANNOT_APPEND            0x87
#define NOT_LOCKED               0x88
#define NO_DESTINATION_TAB       0x89
#define INDEX_OUT_OF_RANGE       0x8a
#define NOT_A_RECORD             0x8b
#define EMPTY                    0x8c
#define NO_OBJECT                0x8d     // not used
#define BAD_PASSWORD             0x8e
#define PTR_TO_DELETED           0x8f
#define NIL_PTR                  0x90
#define OUT_OF_KERNEL_MEMORY     0x91
#define TOO_MANY_CURSORS         0x92     // not used
#define IS_DELETED               0x93
#define INDEX_NOT_FOUND          0x94
#define OBJECT_NOT_FOUND         0x95
#define OUT_OF_APPL_MEMORY       0x96
#define BAD_DATA_SIZE            0x97
#define UNREADABLE_BLOCK         0x98
#define DEMO_VERSION             0x99
#define OUT_OF_BLOCKS            0x9a
#define REQUEST_BREAKED          0x9b
#define OS_FILE_ERROR            0x9d
#define INCOMPATIBLE_VERSION     0x9e
#define REJECTED_BY_KERNEL       0x9f
#define MUST_NOT_BE_NULL         0xa0  // 40002
#define NO_BITAB                 0xa1
#define NO_MY_RIGHT              0xa2
#define IE_OUT_OF_DWORM          0xa3  /* interni chyba */
#define IE_FRAME_OVERRUN         0xa4  /* interni chyba */
#define IE_PAGING                0xa5  /* interni chyba */
#define IE_DOUBLE_PAGE           0xa6  /* interni chyba */
#define IE_OUT_OF_BSTACK         0xa7  /* interni chyba */
#define TABLE_DAMAGED            0xa8
#define CANNOT_LOCK_KERNEL       0xa9
#define END_OF_VOLUME_SPACE      0xaa  /* interni zprava */
#define DEADLOCK                 0xab
#define KEY_DUPLICITY            0xac   // 40002
#define BAD_VERSION              0xad   // client/server version mismatch
#define CHECK_CONSTRAIN          0xae   // 40002
#define REFERENTIAL_CONSTRAIN    0xaf   // 40002
#define UNPROPER_TYPE            0xb0
#define OUTER_JOIN_TOO_COMPLEX   0xb1   // not used
#define REQUEST_NESTING          0xb2
#define CANNOT_FOR_ODBC          0xb3
#define ERROR_IN_FUNCTION_ARG    0xb4
#define ODBC_CURSOR_NOT_OPEN     0xb5
#define DRIVER_NOT_CAPABLE       0xb6
#define TOO_COMPLEX_TRANS        0xb7
#define INTERNAL_SIGNAL          0xb8
#define PRIV_KEY_NOT_FOUND       0xb9
#define NO_WRITE_TOKEN           0xba
#define WAITING_FOR_ACKN         0xbb
#define REPL_BLOCKED             0xbc
#define BAD_TOKEN_STATE          0xbd
#define BAD_TABLE_PROPERTIES     0xbe
#define INDEX_DAMAGED            0xbf
#define PASSWORD_EXPIRED         0xc0
#define NO_KEY_FOUND             0xc1
#define DIFFERENT_KEY            0xc2
#define ASSERTION_FAILED         0xc3
#define SQ_INVALID_CURSOR_STATE  0xc4 // 24000
#define SQ_SAVEPOINT_INVAL_SPEC  0xc5 // 3B001
#define SQ_SAVEPOINT_TOO_MANY    0xc6 // 3B002
#define SQ_TRANS_STATE_ACTIVE    0xc7 // 25001
#define SQ_INVAL_TRANS_TERM      0xc8 // 2D000
#define SQ_TRANS_STATE_RDONLY    0xc9 // 25006
#define SQ_NUM_VAL_OUT_OF_RANGE  0xca // 22003
#define SQ_INV_CHAR_VAL_FOR_CAST 0xcb // 22018
#define SQ_STRING_DATA_RIGHT_TRU 0xcc // 22001
#define SQ_DIVISION_BY_ZERO      0xcd // 22012
#define SQ_CARDINALITY_VIOLATION 0xce // 21000
#define SQ_INVALID_ESCAPE_CHAR   0xcf // 22019
#define SQ_CASE_NOT_FOUND_STMT   0xd0 // 20000
#define SQ_UNHANDLED_USER_EXCEPT 0xd1 // 45000
#define SQ_RESIGNAL_HND_NOT_ACT  0xd2 // 0K000
#define SQ_EXT_ROUT_NOT_AVAIL    0xd3 // 38001
#define SQ_NO_RETURN_IN_FNC      0xd4 // 2F001
#define COLUMN_NOT_EDITABLE      0xd5
#define SQ_TRIGGERED_ACTION      0xd6 // 09000
#define REPLICATION_NOT_RUNNING  0xd7
#define REPL_MAIL_ERROR          0xd8
#define SQ_INVALID_CURSOR_NAME   0xd9 // 34000
#define ROLE_FROM_DIFF_APPL      0xda
#define SEQUENCE_EXHAUSTED       0xdb
#define NO_CURRENT_VAL           0xdc
#define NO_WWW_LICENCE           0xdd
#define NO_MORE_INTRANET_LICS    0xde
#define LIBRARY_ACCESS_DISABLED  0xdf
#define LIBRARY_NOT_FOUND        0xe0

#define LAST_DB_ERROR LIBRARY_NOT_FOUND

#define FIRST_MAIL_ERROR         500  // client mail error numbers
#define MAIL_NOT_INITIALIZED     500
#define MAIL_ERROR               501
#define MAIL_NOT_REMOTE          502
#define MAIL_TYPE_INVALID        503
#define MAIL_LOGON_FAILED        504
#define MAIL_BAD_PROFILE         505
#define MAIL_BAD_USERID          506
#define MAIL_NO_ADDRESS          507
#define FILE_NOT_FOUND           508
#define MAIL_SYSTEM_ACCOUNT      509
#define MAIL_DIAL_ERROR          510
#define MAIL_ALREADY_INIT        511
#define MAIL_PROFILE_NOTFND      512
#define MAIL_PROFSTR_NOTFND      513
#define MAIL_INVALIDPATH         514
#define MAIL_SOCK_IO_ERROR       515
#define MAIL_UNKNOWN_SERVER      516
#define MAIL_CONNECT_FAILED      517
#define MAIL_NO_MORE_FILES       518
#define MAIL_FILE_DELETED        519
#define MAIL_BOX_LOCKED          520
#define MAIL_MSG_NOT_FOUND       521
#define MAIL_UNK_MSG_FMT         522
#define MAIL_DLL_NOT_FOUND       523
#define MAIL_FUNC_NOT_FOUND      524
#define MAIL_NO_SUPPORT          525

/**************************** Cisla varovani: *******************************/
#define NO_WARNING         0
#define WAS_IN_TRANS       1
#define NOT_IN_TRANS       2
#define ERROR_IN_CONSTRS   3
#define IS_NOT_DEL         4
#define ERROR_IN_DEFVAL    5
#define IS_DEL             8
#define IS_EMPTY          16
#define NO_BIPTR          32
#define INDEX_OOR         64
#define DUPL_KEY          96
#define internal_IS_ERROR  128 // hodnota Sz_warning, kdyz doslo k chybe

/************ Hodnoty vracene funkci link_kernel a interf_init **************/
#define KSE_OK               0   /* bez chyby */
#define KSE_WINDOWS          1   /* zahlcene Windows */
#define KSE_INSTAL_PRESENT   2   /* pracuje program WBINSTAL */
#define KSE_ZALOHA_PRESENT   3   /* pracuje program ZALOHA */
#define KSE_WINEXEC          4   /* nelze spustit WBADMIN */
#define KSE_NO_MEMORY        5   /* neni dost pameti */
#define KSE_NO_CONFIG        6   /* nenalezen kofiguracni soubor WB.CNF */
#define KSE_NO_FIL           7   /* jmeno serveru neni registrovano */
#define KSE_DAMAGED          8   /* databaze je vazne poskozena */
#define KSE_NO_ACCESS        9   /* zahlceno databazove jadro */
#define KSE_SERVER_CLOSED    10  /* server zamcen */
#define KSE_BAD_VERSION      11  /* chybna verze jadra */
#define KSE_NETWORK_INIT     12  /* nedari se inicializovat sit */
#define KSE_QUIT             13  /* nevybran zadny server */
#define KSE_NOSERVER         14  /* nenalezen zadny server */
#define KSE_CONNECTION       15  /* nenavazano spojeni se serverem */
#define KSE_NOTASK           16  /* nelze vytvorit dalsi instanci */
#define KSE_WBNETMEM         17  /* nelze spustit program WBNETMEM */
#define KSE_MAXCLIENTS       18  /* prekrocen maximalni pocet clientu */
#define KSE_SPXNOTINSTALLED  19  /* zavedeny ovladac IPX nepodporuje SPX, pravdepodobne pouzito IPXODI /a */
#define KSE_MAXCONNECTION    20  /* neni dost relaci SPX (connections), nastavte v shell.cfg */
#define KSE_MAXSESSION       21  /* neni dost relaci NetBIOS (sessions), viz instalacni manual */
#define KSE_MAXNCB           22  /* neni dost ridicich bloku NetBIOS, viz instalacni manual */
#define KSE_DBASE_OPEN       23  /* databaze je jiz otevrena jinym serverem */
#define KSE_SERVER_NAME_USED 24  /* v siti jiz bezi databazovy server stejneho jmena */
#define KSE_START_THREAD     25  /* nelze spustit dalsi vlakno */
#define KSE_SYNCHRO_OBJ      26  /* nelze vytvorit synchronizacni objekt */
#define KSE_MAPPING          27  /* nelze mapovat pametovy soubor */
#define KSE_CREATE_THREAD    28  /* nelze vytvorit vlakno pro obsluhu klienta */
#define KSE_TIMEOUT          29  /* primo dostupny server neodpovida */
#define KSE_NO_WINSOCK       30  /* knihovna winsock neni instalovana */
#define KSE_WINSOCK_ERROR    31  /* chyba pri praci se socketem */
#define KSE_SERVER_NOT_KNOWN 32  /* jmeno serveru neni registrovano */
#define KSE_SERVER_PATH_NOT_SPECIF 33 /* cesta k databazi serveru neni registrovana */
#define KSE_NETBIOS_NAME     34  /* nemohu pridat jmeno pro NetBIOS */
#define KSE_FWNOTFOUND       35  /* nenalezen firewall */
#define KSE_FWCOMM           36  /* nelze komunikovat s firewallem */
#define KSE_FWDENIED         37  /* firewall odmitl vytvorit propojeni */
#define KSE_BAD_PASSWORD     38  /* chyba hesla k databazovemu souboru */
#define KSE_ESCAPED          39  /* upusteno od startu serveru */
#define KSE_NO_IPX           40  /* protokol IPX neni k dispozici */
#define KSE_EXPIRED          41  /* skoncila pouzitelnost serveru */
#define KSE_CANNOT_CREATE_FIL 42 /* nelze vytvorit databazovy soubor */
#define KSE_CANNOT_OPEN_FIL   43 /* nelze otevrit databazovy soubor */
#define KSE_CANNOT_OPEN_TRANS 44 /* nelze otevrit transakcni soubor */
#define KSE_IP_FILTER         45 /* pristup z teto IP adresy neni povolen */
#define KSE_NO_SERVER_LIST    46 /* nenalezen soubor wbcgi.conf se seznamem serveru */
#define KSE_LAST            47

/******* Typy atributu (vyuziji se pri volani funkce Enum_attributes) *******/
#define ATT_BOOLEAN    1            /* Boolean */
#define ATT_CHAR       2            /* Char */
#define ATT_INT16      3            /* Short */
#define ATT_INT32      4            /* Integer */
#define ATT_MONEY      5            /* Money */
#define ATT_FLOAT      6            /* Real */
#define ATT_STRING     7            /* String + size*/
#define ATT_CSSTRING   8            /* CSString + size */
#define ATT_CSISTRING  9            /* CSIString + size*/
#define ATT_BINARY    10            /* Binary + size */
#define ATT_DATE      11            /* Date */
#define ATT_TIME      12            /* Time */
#define ATT_TIMESTAMP 13            /* Timestamp */
#define ATT_PTR       14            /* Pointer */
#define ATT_BIPTR     15            /* Bipointer */

#define ATT_FIRSTSPEC      16
#define ATT_AUTOR     16            /* sledovaci atribut: Autorizace */
#define ATT_DATIM     17            /* sledovaci atribut: Datumovka */
#define ATT_HIST      18            /* sledovaci atribut: Historie */
#define ATT_LASTSPEC       18

#define ATT_FIRST_HEAPTYPE 18
#define ATT_RASTER    19            /* Raster */
#define ATT_TEXT      20            /* Text */
#define ATT_NOSPEC    21            /* Nospec */
#define ATT_SIGNAT    22            /* Signature */
#define ATT_LAST_HEAPTYPE  22

#endif   /* !def __GENERAL_H__ */

